import warnings
from typing import Optional
from uuid import UUID

import rest_framework
from rest_framework import fields, relations
from rest_framework.exceptions import ValidationError

from .action import Actions
from .mixins import (
    ActionMixin, RenderMixin, DisplayMode, NullChoiceMixin, RelatedFieldAJAXMixin, FieldHelpTextMixin,
    DateTimeFieldMixin, DateFieldMixin, TimeFieldMixin, HiddenFieldMixin
)
from .settings import version_check


class BooleanField(RenderMixin, ActionMixin, FieldHelpTextMixin, fields.BooleanField):
    NULL_VALUES = {'null', 'Null', 'NULL', '', None}

    def __init__(self, read_only=False, write_only=False, required=None, default=fields.empty, initial=fields.empty,
                 source=None, label=None, help_text=None, style=None, error_messages=None, validators=None,
                 actions: Actions = None, uuid: UUID = None, display: DisplayMode = None,
                 display_table: DisplayMode = None, display_form: DisplayMode = None, table_classes: str = '', **kw):
        kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
        kwargs.update(kw)
        super().__init__(**kwargs)

    def to_internal_value(self, data):
        try:
            parent_value = super().to_internal_value(data)
            return parent_value
        except TypeError:  # Input is an unhashable type
            pass
        except ValidationError:
            if data in self.NULL_VALUES and self.allow_null:
                return None
        self.fail('invalid', input=data)

    def to_representation(self, value):
        parent_value = super().to_representation(value)
        if value in self.NULL_VALUES and self.allow_null:
            return None
        return parent_value


class NullBooleanField(RenderMixin, ActionMixin, FieldHelpTextMixin, fields.NullBooleanField):

    def __init__(self, read_only=False, write_only=False, required=None, default=fields.empty, initial=fields.empty,
                 source=None, label=None, help_text=None, style=None, error_messages=None, validators=None,
                 actions: Actions = None, uuid: UUID = None, display: DisplayMode = None,
                 display_table: DisplayMode = None, display_form: DisplayMode = None, table_classes: str = '', **kw):
        kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
        kwargs.update(kw)
        super().__init__(**kwargs)


class CharField(RenderMixin, ActionMixin, FieldHelpTextMixin, fields.CharField):

    def __init__(self, read_only=False, write_only=False, required=None, default=fields.empty, initial=fields.empty,
                 source=None, label=None, help_text=None, style=None, error_messages=None, validators=None,
                 allow_null=False, actions: Actions = None, uuid: UUID = None, display: DisplayMode = None,
                 display_table: DisplayMode = None, display_form: DisplayMode = None, table_classes: str = '', **kw):
        kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
        kwargs.update(kw)
        super().__init__(**kwargs)


class EmailField(RenderMixin, ActionMixin, FieldHelpTextMixin, fields.EmailField):

    def __init__(self, read_only=False, write_only=False, required=None, default=fields.empty, initial=fields.empty,
                 source=None, label=None, help_text=None, style=None, error_messages=None, validators=None,
                 allow_null=False, actions: Actions = None, uuid: UUID = None, display: DisplayMode = None,
                 display_table: DisplayMode = None, display_form: DisplayMode = None, table_classes: str = '', **kw):
        kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
        kwargs.update(kw)
        super().__init__(**kwargs)


class RegexField(RenderMixin, ActionMixin, FieldHelpTextMixin, fields.RegexField):

    def __init__(self, regex, read_only=False, write_only=False, required=None, default=fields.empty,
                 initial=fields.empty, source=None, label=None, help_text=None, style=None, error_messages=None,
                 validators=None, allow_null=False, actions: Actions = None, uuid: UUID = None,
                 display: DisplayMode = None, display_table: DisplayMode = None, display_form: DisplayMode = None,
                 table_classes: str = '', **kw):
        kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
        kwargs.update(kw)
        super().__init__(**kwargs)


class SlugField(RenderMixin, ActionMixin, FieldHelpTextMixin, fields.SlugField):

    def __init__(self, allow_unicode=False, read_only=False, write_only=False, required=None, default=fields.empty,
                 initial=fields.empty, source=None, label=None, help_text=None, style=None, error_messages=None,
                 validators=None, allow_null=False, actions: Actions = None, uuid: UUID = None,
                 display: DisplayMode = None, display_table: DisplayMode = None, display_form: DisplayMode = None,
                 table_classes: str = '', **kw):
        kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
        kwargs.update(kw)
        # noinspection PyUnresolvedReferences
        if not version_check(rest_framework.VERSION, '3.6.4'):
            kwargs.pop('allow_unicode', None)
        super().__init__(**kwargs)


class URLField(RenderMixin, ActionMixin, FieldHelpTextMixin, fields.URLField):

    def __init__(self, read_only=False, write_only=False, required=None, default=fields.empty, initial=fields.empty,
                 source=None, label=None, help_text=None, style=None, error_messages=None, validators=None,
                 allow_null=False, actions: Actions = None, uuid: UUID = None, display: DisplayMode = None,
                 display_table: DisplayMode = None, display_form: DisplayMode = None, table_classes: str = '', **kw):
        kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
        kwargs.update(kw)
        super().__init__(**kwargs)


class UUIDField(RenderMixin, ActionMixin, FieldHelpTextMixin, fields.UUIDField):

    def __init__(self, read_only=False, write_only=False, required=None, default=fields.empty, initial=fields.empty,
                 source=None, label=None, help_text=None, style=None, error_messages=None, validators=None,
                 allow_null=False, actions: Actions = None, uuid: UUID = None, display: DisplayMode = None,
                 display_table: DisplayMode = None, display_form: DisplayMode = None, table_classes: str = '', **kw):
        kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
        kwargs.update(kw)
        super().__init__(**kwargs)


class IPAddressField(RenderMixin, ActionMixin, FieldHelpTextMixin, fields.IPAddressField):

    def __init__(self, protocol='both', read_only=False, write_only=False, required=None, default=fields.empty,
                 initial=fields.empty, source=None, label=None, help_text=None, style=None, error_messages=None,
                 validators=None, allow_null=False, actions: Actions = None, uuid: UUID = None,
                 display: DisplayMode = None, display_table: DisplayMode = None, display_form: DisplayMode = None,
                 table_classes: str = '', **kw):
        kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
        kwargs.update(kw)
        super().__init__(**kwargs)


class IntegerField(RenderMixin, ActionMixin, FieldHelpTextMixin, fields.IntegerField):

    def __init__(self, read_only=False, write_only=False, required=None, default=fields.empty, initial=fields.empty,
                 source=None, label=None, help_text=None, style=None, error_messages=None, validators=None,
                 allow_null=False, actions: Actions = None, uuid: UUID = None, display: DisplayMode = None,
                 display_table: DisplayMode = None, display_form: DisplayMode = None, table_classes: str = '', **kw):
        kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
        kwargs.update(kw)
        super().__init__(**kwargs)


class FloatField(RenderMixin, ActionMixin, FieldHelpTextMixin, fields.FloatField):

    def __init__(self, read_only=False, write_only=False, required=None, default=fields.empty, initial=fields.empty,
                 source=None, label=None, help_text=None, style=None, error_messages=None, validators=None,
                 allow_null=False, actions: Actions = None, uuid: UUID = None, display: DisplayMode = None,
                 display_table: DisplayMode = None, display_form: DisplayMode = None, table_classes: str = '', **kw):
        kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
        kwargs.update(kw)
        super().__init__(**kwargs)


class DecimalField(RenderMixin, ActionMixin, FieldHelpTextMixin, fields.DecimalField):

    def __init__(self, max_digits, decimal_places, coerce_to_string=None, max_value=None, min_value=None,
                 localize=False, rounding=None, read_only=False, write_only=False, required=None, default=fields.empty,
                 initial=fields.empty, source=None, label=None, help_text=None, style=None, error_messages=None,
                 validators=None, allow_null=False, actions: Actions = None, uuid: UUID = None,
                 display: DisplayMode = None, display_table: DisplayMode = None, display_form: DisplayMode = None,
                 table_classes: str = '', **kw):
        kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
        kwargs.update(kw)
        # noinspection PyUnresolvedReferences
        if not version_check(rest_framework.VERSION, '3.7.2'):
            kwargs.pop('rounding', None)
        # noinspection PyUnresolvedReferences
        if not version_check(rest_framework.VERSION, '3.4.0'):
            kwargs.pop('localize', None)
        super().__init__(**kwargs)


# noinspection PyShadowingBuiltins
class DateTimeField(DateTimeFieldMixin, RenderMixin, ActionMixin, FieldHelpTextMixin, fields.DateTimeField):

    def __init__(self, format=fields.empty, input_formats=None, default_timezone=None, read_only=False,
                 write_only=False, required=None, default=fields.empty, initial=fields.empty, source=None, label=None,
                 help_text=None, style=None, error_messages=None, validators=None, allow_null=False,
                 actions: Actions = None, uuid: UUID = None, display: DisplayMode = None,
                 display_table: DisplayMode = None, display_form: DisplayMode = None, table_classes: str = '', **kw):
        kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
        kwargs.update(kw)
        super().__init__(**kwargs)


# noinspection PyShadowingBuiltins
class DateField(DateFieldMixin, RenderMixin, ActionMixin, FieldHelpTextMixin, fields.DateField):

    def __init__(self, format=fields.empty, input_formats=None, read_only=False, write_only=False, required=None,
                 default=fields.empty, initial=fields.empty, source=None, label=None, help_text=None, style=None,
                 error_messages=None, validators=None, allow_null=False, actions: Actions = None, uuid: UUID = None,
                 display: DisplayMode = None, display_table: DisplayMode = None, display_form: DisplayMode = None,
                 table_classes: str = '', **kw):
        self.time_step = kw.pop('time_step', None)
        kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
        kwargs.update(kw)
        super().__init__(**kwargs)


# noinspection PyShadowingBuiltins
class TimeField(TimeFieldMixin, RenderMixin, ActionMixin, FieldHelpTextMixin, fields.TimeField):

    def __init__(self, format=fields.empty, input_formats=None, read_only=False, write_only=False, required=None,
                 default=fields.empty, initial=fields.empty, source=None, label=None, help_text=None, style=None,
                 error_messages=None, validators=None, allow_null=False, actions: Actions = None, uuid: UUID = None,
                 display: DisplayMode = None, display_table: DisplayMode = None, display_form: DisplayMode = None,
                 table_classes: str = '', **kw):
        self.time_step = kw.pop('time_step', None)
        kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
        kwargs.update(kw)
        super().__init__(**kwargs)


class DurationField(RenderMixin, ActionMixin, FieldHelpTextMixin, fields.DurationField):

    def __init__(self, read_only=False, write_only=False, required=None, default=fields.empty, initial=fields.empty,
                 source=None, label=None, help_text=None, style=None, error_messages=None, validators=None,
                 allow_null=False, actions: Actions = None, uuid: UUID = None, display: DisplayMode = None,
                 display_table: DisplayMode = None, display_form: DisplayMode = None, table_classes: str = '', **kw):
        kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
        kwargs.update(kw)
        super().__init__(**kwargs)


class ChoiceField(NullChoiceMixin, RenderMixin, ActionMixin, FieldHelpTextMixin, fields.ChoiceField):

    def __init__(self, choices, read_only=False, write_only=False, required=None, default=fields.empty,
                 initial=fields.empty, source=None, label=None, help_text=None, style=None, error_messages=None,
                 validators=None, allow_null=False, actions: Actions = None, uuid: UUID = None,
                 display: DisplayMode = None, display_table: DisplayMode = None, display_form: DisplayMode = None,
                 table_classes: str = '', **kw):
        kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
        kwargs.update(kw)
        super().__init__(**kwargs)


class MultipleChoiceField(NullChoiceMixin, RenderMixin, ActionMixin, FieldHelpTextMixin, fields.MultipleChoiceField):

    def __init__(self, choices, read_only=False, write_only=False, required=None, default=fields.empty,
                 initial=fields.empty, source=None, label=None, help_text=None, style=None, error_messages=None,
                 validators=None, allow_null=False, actions: Actions = None, uuid: UUID = None,
                 display: DisplayMode = None, display_table: DisplayMode = None, display_form: DisplayMode = None,
                 table_classes: str = '', **kw):
        kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
        kwargs.update(kw)
        super().__init__(**kwargs)


class FilePathField(NullChoiceMixin, RenderMixin, ActionMixin, FieldHelpTextMixin, fields.FilePathField):

    def __init__(self, path, match=None, recursive=False, allow_files=True, allow_folders=False, required=None,
                 read_only=False, write_only=False, default=fields.empty, initial=fields.empty, source=None,
                 label=None, help_text=None, style=None, error_messages=None, validators=None, allow_null=False,
                 actions: Actions = None, uuid: UUID = None, display: DisplayMode = None,
                 display_table: DisplayMode = None, display_form: DisplayMode = None, table_classes: str = '', **kw):
        kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
        kwargs.update(kw)
        super().__init__(**kwargs)


class FileField(RenderMixin, ActionMixin, FieldHelpTextMixin, fields.FileField):

    def __init__(self, read_only=False, write_only=False, required=None, default=fields.empty, initial=fields.empty,
                 source=None, label=None, help_text=None, style=None, error_messages=None, validators=None,
                 allow_null=False, actions: Actions = None, uuid: UUID = None, display: DisplayMode = None,
                 display_table: DisplayMode = None, display_form: DisplayMode = None, table_classes: str = '', **kw):
        kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
        kwargs.update(kw)
        super().__init__(**kwargs)


class ImageField(RenderMixin, ActionMixin, FieldHelpTextMixin, fields.ImageField):

    def __init__(self, read_only=False, write_only=False, required=None, default=fields.empty, initial=fields.empty,
                 source=None, label=None, help_text=None, style=None, error_messages=None, validators=None,
                 allow_null=False, actions: Actions = None, uuid: UUID = None, display: DisplayMode = None,
                 display_table: DisplayMode = None, display_form: DisplayMode = None, table_classes: str = '', **kw):
        kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
        kwargs.update(kw)
        super().__init__(**kwargs)


class ListField(RenderMixin, ActionMixin, FieldHelpTextMixin, fields.ListField):

    def __init__(self, read_only=False, write_only=False, required=None, default=fields.empty, initial=fields.empty,
                 source=None, label=None, help_text=None, style=None, error_messages=None, validators=None,
                 allow_null=False, actions: Actions = None, uuid: UUID = None, display: DisplayMode = None,
                 display_table: DisplayMode = None, display_form: DisplayMode = None, table_classes: str = '', **kw):
        kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
        kwargs.update(kw)
        super().__init__(**kwargs)


class DictField(RenderMixin, ActionMixin, FieldHelpTextMixin, fields.DictField):

    def __init__(self, read_only=False, write_only=False, required=None, default=fields.empty, initial=fields.empty,
                 source=None, label=None, help_text=None, style=None, error_messages=None, validators=None,
                 allow_null=False, actions: Actions = None, uuid: UUID = None, display: DisplayMode = None,
                 display_table: DisplayMode = None, display_form: DisplayMode = None, table_classes: str = '', **kw):
        kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
        kwargs.update(kw)
        super().__init__(**kwargs)


if hasattr(fields, 'HStoreField'):
    class HStoreField(RenderMixin, ActionMixin, FieldHelpTextMixin, fields.HStoreField):

        def __init__(self, read_only=False, write_only=False, required=None, default=fields.empty,
                     initial=fields.empty, source=None, label=None, help_text=None, style=None, error_messages=None,
                     validators=None, allow_null=False, actions: Actions = None, uuid: UUID = None,
                     display: DisplayMode = None, display_table: DisplayMode = None, display_form: DisplayMode = None,
                     table_classes: str = '', **kw):
            kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
            kwargs.update(kw)
            super().__init__(**kwargs)


class JSONField(RenderMixin, ActionMixin, FieldHelpTextMixin, fields.JSONField):

    def __init__(self, read_only=False, write_only=False, required=None, default=fields.empty, initial=fields.empty,
                 source=None, label=None, help_text=None, style=None, error_messages=None, validators=None,
                 allow_null=False, actions: Actions = None, uuid: UUID = None, display: DisplayMode = None,
                 display_table: DisplayMode = None, display_form: DisplayMode = None, table_classes: str = '', **kw):
        kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
        kwargs.update(kw)
        super().__init__(**kwargs)


# noinspection PyAbstractClass
class ReadOnlyField(RenderMixin, ActionMixin, FieldHelpTextMixin, fields.ReadOnlyField):

    def __init__(self, read_only=False, write_only=False, required=None, default=fields.empty, initial=fields.empty,
                 source=None, label=None, help_text=None, style=None, error_messages=None, validators=None,
                 allow_null=False, actions: Actions = None, uuid: UUID = None, display: DisplayMode = None,
                 display_table: DisplayMode = None, display_form: DisplayMode = None, table_classes: str = '', **kw):
        warnings.warn('deprecated - wrong approach! Use read_only attribute instead.',
                      DeprecationWarning, stacklevel=2)
        read_only = True  # NOQA
        kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
        kwargs.update(kw)
        super().__init__(**kwargs)


# noinspection PyAbstractClass
class HiddenField(HiddenFieldMixin, RenderMixin, ActionMixin, FieldHelpTextMixin, fields.HiddenField):

    def __init__(self, read_only=False, write_only=False, required=None, default=fields.empty, initial=fields.empty,
                 source=None, label=None, help_text=None, style=None, error_messages=None, validators=None,
                 allow_null=False, actions: Actions = None, uuid: UUID = None, display: DisplayMode = None,
                 display_table: DisplayMode = None, display_form: DisplayMode = None, table_classes: str = '', **kw):
        warnings.warn('deprecated - wrong approach! Use display(|_table|_form) attributes instead.',
                      DeprecationWarning, stacklevel=2)
        display = DisplayMode.HIDDEN  # NOQA
        kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
        kwargs.update(kw)
        super().__init__(**kwargs)


# noinspection PyAbstractClass
class SerializerMethodField(RenderMixin, ActionMixin, FieldHelpTextMixin, fields.SerializerMethodField):

    def __init__(self, method_name=None, read_only=False, write_only=False, required=None, default=fields.empty,
                 initial=fields.empty, source=None, label=None, help_text=None, style=None, error_messages=None,
                 validators=None, allow_null=False, actions: Actions = None, uuid: UUID = None,
                 display: DisplayMode = None, display_table: DisplayMode = None, display_form: DisplayMode = None,
                 table_classes: str = '', **kw):
        kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
        kwargs.update(kw)
        super().__init__(**kwargs)


class ModelField(RenderMixin, ActionMixin, FieldHelpTextMixin, fields.ModelField):

    def __init__(self, model_field, read_only=False, write_only=False, required=None, default=fields.empty,
                 initial=fields.empty, source=None, label=None, help_text=None, style=None, error_messages=None,
                 validators=None, allow_null=False, actions: Actions = None, uuid: UUID = None,
                 display: DisplayMode = None, display_table: DisplayMode = None, display_form: DisplayMode = None,
                 table_classes: str = '', **kw):
        kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
        kwargs.update(kw)
        super().__init__(**kwargs)


class StringRelatedField(RelatedFieldAJAXMixin, RenderMixin, ActionMixin, FieldHelpTextMixin,
                         relations.StringRelatedField):
    def __init__(self, read_only=False, write_only=False, required=None, default=fields.empty, initial=fields.empty,
                 source=None, label=None, help_text=None, style=None, error_messages=None, validators=None,
                 allow_null=False, url_reverse: Optional[str] = None, placeholder: Optional[str] = None,
                 additional_parameters: Optional[dict] = None, query_field: str = 'query', actions: Actions = None,
                 uuid: UUID = None, display: DisplayMode = None, display_table: DisplayMode = None,
                 display_form: DisplayMode = None, table_classes: str = '', **kw):
        kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
        kwargs.update(kw)
        super().__init__(**kwargs)


class PrimaryKeyRelatedField(RelatedFieldAJAXMixin, RenderMixin, ActionMixin, FieldHelpTextMixin,
                             relations.PrimaryKeyRelatedField):
    def __init__(self, read_only=False, write_only=False, required=None, default=fields.empty, initial=fields.empty,
                 source=None, label=None, help_text=None, style=None, error_messages=None, validators=None,
                 allow_null=False, url_reverse: Optional[str] = None, placeholder: Optional[str] = None,
                 additional_parameters: Optional[dict] = None, query_field: str = 'query', actions: Actions = None,
                 uuid: UUID = None, display: DisplayMode = None, display_table: DisplayMode = None,
                 display_form: DisplayMode = None, table_classes: str = '', **kw):
        kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
        kwargs.update(kw)
        super().__init__(**kwargs)


class HyperlinkedRelatedField(RelatedFieldAJAXMixin, RenderMixin, ActionMixin, FieldHelpTextMixin,
                              relations.HyperlinkedRelatedField):
    def __init__(self, view_name=None, read_only=False, write_only=False, required=None, default=fields.empty,
                 initial=fields.empty, source=None, label=None, help_text=None, style=None, error_messages=None,
                 validators=None, allow_null=False, url_reverse: Optional[str] = None,
                 placeholder: Optional[str] = None, additional_parameters: Optional[dict] = None,
                 query_field: str = 'query', actions: Actions = None, uuid: UUID = None, display: DisplayMode = None,
                 display_table: DisplayMode = None, display_form: DisplayMode = None, table_classes: str = '', **kw):
        kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
        kwargs.update(kw)
        super().__init__(**kwargs)


class HyperlinkedIdentityField(RelatedFieldAJAXMixin, RenderMixin, ActionMixin, FieldHelpTextMixin,
                               relations.HyperlinkedIdentityField):
    def __init__(self, view_name=None, read_only=False, write_only=False, required=None, default=fields.empty,
                 initial=fields.empty, source=None, label=None, help_text=None, style=None, error_messages=None,
                 validators=None, allow_null=False, url_reverse: Optional[str] = None,
                 placeholder: Optional[str] = None, additional_parameters: Optional[dict] = None,
                 query_field: str = 'query', actions: Actions = None, uuid: UUID = None, display: DisplayMode = None,
                 display_table: DisplayMode = None, display_form: DisplayMode = None, table_classes: str = '', **kw):
        kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
        kwargs.update(kw)
        super().__init__(**kwargs)


class SlugRelatedField(RelatedFieldAJAXMixin, RenderMixin, ActionMixin, FieldHelpTextMixin, relations.SlugRelatedField):

    def __init__(self, slug_field=None, read_only=False, write_only=False, required=None, default=fields.empty,
                 initial=fields.empty, source=None, label=None, help_text=None, style=None, error_messages=None,
                 validators=None, allow_null=False, url_reverse: Optional[str] = None,
                 placeholder: Optional[str] = None, additional_parameters: Optional[dict] = None,
                 query_field: str = 'query', actions: Actions = None, uuid: UUID = None, display: DisplayMode = None,
                 display_table: DisplayMode = None, display_form: DisplayMode = None, table_classes: str = '', **kw):
        kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
        kwargs.update(kw)
        super().__init__(**kwargs)


class ManyRelatedField(RenderMixin, ActionMixin, FieldHelpTextMixin, relations.ManyRelatedField):

    def __init__(self, child_relation=None, read_only=False, write_only=False, required=None, default=fields.empty,
                 initial=fields.empty, source=None, label=None, help_text=None, style=None, error_messages=None,
                 validators=None, allow_null=True, actions: Actions = None, uuid: UUID = None,
                 display: DisplayMode = None, display_table: DisplayMode = None, display_form: DisplayMode = None,
                 table_classes: str = '', **kw):
        kwargs = {k: v for k, v in locals().items() if not k.startswith(('__', 'self', 'kw'))}
        kwargs.update(kw)
        super().__init__(**kwargs)



import uuid as uuid_module
from enum import IntEnum
from typing import Iterable, List, Union

from django.utils.translation import ugettext_lazy as _
from rest_framework.serializers import Serializer

from .settings import DYNAMICFORMS


class ActionBase(object):

    def __init__(self, action_js: str, name: Union[str, None] = None, serializer: Serializer = None):
        """
        :param action_js: JavaScript to execute when action is run
        :param name: name by which to recognise this action in further processing, e.g. Serializer.suppress_action
        :param serializer: bind to this serializer instance
        """
        self.name = name
        self.action_js = action_js
        assert self.action_js is not None, 'When declaring action, it must declare action JavaScript to execute'
        # serializer will be set when obtaining a resolved copy
        self.serializer = serializer

    @property
    def action_id(self):
        return id(self)

    def copy_and_resolve_reference(self, serializer):
        raise NotImplementedError()

    def render(self, serializer: Serializer, **kwds):
        raise NotImplementedError()

    @staticmethod
    def prepare_string(string, encode=True):
        """
        Replaces curly brackets with some special string so there is no problem when string.format() is called

        :param string: String in which program searches for curly brackets
        :param encode: True: replaces brackets with special string, False: the other way around
        :return:
        """
        if not string:
            return ''
        replaces = {'{': '´|curl_brack_start|`', '}': '´|curl_brack_end|`'}
        if not encode:
            replaces = dict([(value, key) for key, value in replaces.items()])

        for key, val in replaces.items():
            string = string.replace(key, val)
        return string


class RenderableActionMixin(object):
    """
    Action that is rendered on screen
    """

    def __init__(self, label: str, title: str, icon: str = None, btn_classes: Union[str, dict, None] = None):
        """
        :param label: Label for rendering to on screen control
        :param title: Hint text for on-screen control
        :param icon: optional icon to add to render
        :param btn_classes: optional class(es) of button. if variable is dict and key 'replace' have value True,
         then default class will be replaced with class(es) that are under key 'classes'. In other case class(es) will
         be just added to default class
        """
        self.label = label
        self.title = title
        self.icon = icon
        self.btn_classes = btn_classes


class TablePosition(IntEnum):
    HEADER = 0  # Table header of list view
    # On left click on table row (currently this renders only once per table.
    # We might need to add one that renders for each row
    ROW_CLICK = 1
    ROW_RIGHTCLICK = 2  # On right click on table row
    ROW_START = 3  # Additional control column on left side of table
    ROW_END = 4  # Additional control column on right side of table
    FIELD_START = 5  # On left side of field value
    FIELD_END = 6  # On right side of field value


class TableAction(ActionBase, RenderableActionMixin):

    def __init__(self, position: TablePosition, label: str, action_js: str,
                 title: Union[str, None] = None, icon: Union[str, None] = None, field_name: Union[str, None] = None,
                 name: Union[str, None] = None, serializer: Serializer = None,
                 btn_classes: Union[str, dict, None] = None):
        ActionBase.__init__(self, action_js, name, serializer)
        RenderableActionMixin.__init__(self, label, title, icon, btn_classes)
        self.position = position
        self.field_name = field_name

    def copy_and_resolve_reference(self, serializer: Serializer):
        return TableAction(self.position, self.label, self.action_js, self.title, self.icon, self.field_name, self.name,
                           serializer, self.btn_classes)

    def render(self, serializer: Serializer, **kwds):
        ret = rowclick = rowrclick = ''
        stop_propagation = 'if(event.stopPropagation){event.stopPropagation();}event.cancelBubble=true;'

        action_action = self.prepare_string(self.action_js)
        if self.position != TablePosition.HEADER:
            # We need to do this differently because of dynamic page loading for tables: each time the serializer
            # has a different UUID
            action_action = action_action.replace('__TABLEID__',
                                                  "$(event.target).parents('table').attr('id').substr(5)")
        else:
            action_action = action_action.replace('__TABLEID__', "'" + str(serializer.uuid) + "'")

        if self.position == TablePosition.ROW_CLICK:
            rowclick = action_action
        elif self.position == TablePosition.ROW_RIGHTCLICK:
            rowrclick = action_action
        else:
            from uuid import uuid1

            def get_btn_class(default_class, additional_class):
                classes = default_class.split(' ')
                if additional_class:
                    if isinstance(additional_class, dict):
                        if additional_class.get('replace', False):
                            classes = []
                        additional_class = additional_class.get('classes', '').split(' ')
                    else:
                        additional_class = additional_class.split(' ')
                    classes.extend(additional_class)
                return ' '.join(classes)

            button_name = (' name="btn-%s" ' % self.name) if self.name else ''
            btn_class = get_btn_class("btn btn-info", self.btn_classes)

            btnid = uuid1()
            ret += '<button id="df-action-btn-{btnid}" type="button" class="{btn_class}" title="{title}"{button_name}' \
                   'onClick="{stop_propagation} {action}">{icon_def}{label}</button>'. \
                format(btnid=btnid, stop_propagation=stop_propagation, action=action_action,
                       btn_class=btn_class, label=self.prepare_string(self.label),
                       title=self.prepare_string(self.title),
                       icon_def='<img src="{icon}"/>'.format(icon=self.icon) if self.icon else '',
                       button_name=button_name)
            if DYNAMICFORMS.jquery_ui:
                ret += '<script type="application/javascript">$("#df-action-btn-{btnid}").button();</script>' \
                    .format(btnid=btnid)

        if self.position in (TablePosition.ROW_CLICK, TablePosition.ROW_RIGHTCLICK):
            if rowclick != '':
                ret += "$('#list-{uuid}').find('tbody').click(" \
                       "function(event) {{ \n{stop_propagation} \n{action} \nreturn false;\n}});\n". \
                    format(stop_propagation=stop_propagation, action=rowclick, uuid=serializer.uuid)
            if rowrclick != '':
                ret += "$('#list-{uuid}').find('tbody').contextmenu(" \
                       "function(event) {{ \n{stop_propagation} \n{action} \nreturn false;\n}});\n". \
                    format(stop_propagation=stop_propagation, action=rowrclick, uuid=serializer.uuid)
            if ret != '':
                ret = '<script type="application/javascript">%s</script>' % ret

        elif ret != '':
            ret = '<div class="dynamicforms-actioncontrol float-{direction} pull-{direction}">{ret}</div>'.format(
                ret=ret, direction='left' if self.position == TablePosition.FIELD_START else 'right'
            )

        return self.prepare_string(ret, False)


class FieldChangeAction(ActionBase):

    def __init__(self, tracked_fields: Iterable[str], action_js: str, name: Union[str, None] = None,
                 serializer: Serializer = None):
        super().__init__(action_js, name, serializer)
        self.tracked_fields = tracked_fields
        assert self.tracked_fields, 'When declaring an action, it must track at least one form field'
        if serializer:
            self.tracked_fields = [self._resolve_reference(f) for f in self.tracked_fields]

    def _resolve_reference(self, ref):
        from .mixins import RenderMixin

        if isinstance(ref, uuid_module.UUID):
            return str(ref)
        elif isinstance(ref, RenderMixin):
            # TODO unit tests!!!
            # TODO test what happens if the Field instance given is from another serializer
            # TODO test what happens when Field instance is actually a Serializer (when should onchange trigger for it?)
            return ref.uuid
        elif isinstance(ref, str) and ref in self.serializer.fields:
            return self.serializer.fields[ref].uuid
        elif isinstance(ref, str) and '.' in ref:
            # This supports nested serializers and fields with . notation, e.g. master_serializer_field.child_field
            f = self.serializer
            for r in ref.split('.'):
                f = f.fields[r]
            return f.uuid
        raise Exception('Unknown reference type for Action tracked field (%r)' % ref)

    def copy_and_resolve_reference(self, serializer: Serializer):
        return FieldChangeAction(self.tracked_fields, self.action_js, self.name, serializer)

    def render(self, serializer: Serializer, **kwds):
        res = 'var action_func{0.action_id} = {0.action_js};\n'.format(self)
        for tracked_field in self.tracked_fields:
            res += "dynamicforms.registerFieldAction('{ser.uuid}', '{tracked_field}', action_func{s.action_id});\n" \
                .format(ser=serializer, tracked_field=tracked_field, s=self)
        return res


class FormInitAction(ActionBase):

    def copy_and_resolve_reference(self, serializer: Serializer):
        return FormInitAction(self.action_js, self.name, serializer)

    def render(self, serializer: Serializer, **kwds):
        # we need window.setTimeout because at the time of form generation, the initial fields value collection
        # hasn't been done yet
        return 'window.setTimeout(function() {{ {0.action_js} }}, 1);\n'.format(self)


class FieldInitAction(FieldChangeAction):

    def copy_and_resolve_reference(self, serializer: Serializer):
        return FieldInitAction(self.tracked_fields, self.action_js, self.name, serializer)

    def render(self, serializer: Serializer, **kwds):
        # we need window.setTimeout because at the time of form generation, the initial fields value collection
        # hasn't been done yet
        return 'window.setTimeout(function() {{ {0.action_js} }}, 1);;\n'.format(self)


class FormButtonTypes(IntEnum):
    CANCEL = 1
    SUBMIT = 2
    CUSTOM = 3


class FormButtonAction(ActionBase):
    DEFAULT_LABELS = {
        FormButtonTypes.CANCEL: _('Cancel'),
        FormButtonTypes.SUBMIT: _('Save changes'),
        FormButtonTypes.CUSTOM: _('Custom'),
    }

    def __init__(self, btn_type: FormButtonTypes, label: str = None, btn_classes: str = None, action_js: str = None,
                 button_is_primary: bool = None, positions: List[str] = None,
                 name: Union[str, None] = None, serializer: Serializer = None):
        super().__init__(action_js or False, name, serializer)
        self.uuid = uuid_module.uuid1()
        self.btn_type = btn_type
        self.label = label or FormButtonAction.DEFAULT_LABELS[btn_type or FormButtonTypes.CUSTOM]
        self.positions = positions or ['dialog', 'form']

        if button_is_primary is None:
            button_is_primary = btn_type == FormButtonTypes.SUBMIT
        self.button_is_primary = button_is_primary

        self.btn_classes = btn_classes or (
            DYNAMICFORMS.form_button_classes + ' '
            + (DYNAMICFORMS.form_button_classes_primary if button_is_primary
               else DYNAMICFORMS.form_button_classes_secondary) + ' '
            + (DYNAMICFORMS.form_button_classes_cancel if btn_type == FormButtonTypes.CANCEL else '')
        )

    def copy_and_resolve_reference(self, serializer):
        return FormButtonAction(self.btn_type, self.label, self.btn_classes, self.action_js, self.button_is_primary,
                                self.positions, self.name, serializer)

    def render(self, serializer: Serializer, position=None, **kwds):
        if self.btn_type == FormButtonTypes.CANCEL and position == 'form':
            return ''
        action_js = self.action_js
        button_name = ('name="btn-%s"' % self.name) if self.name else ''
        if isinstance(action_js, str):
            action_js = self.prepare_string(action_js)
            action_js = action_js.format(**locals())
        button_type = 'button' if self.btn_type != FormButtonTypes.SUBMIT or position == 'dialog' else 'submit'
        data_dismiss = 'data-dismiss="modal"' if self.btn_type == FormButtonTypes.CANCEL else ''
        if self.btn_type == FormButtonTypes.SUBMIT:
            button_id = 'save-' + str(serializer.uuid)
        else:
            button_id = 'formbutton-' + str(self.uuid)
        if (self.btn_type != FormButtonTypes.SUBMIT or position == 'dialog') and action_js:
            button_js = ('<script type="text/javascript">'
                         '  $("#{button_id}").on("click", function() {{'
                         '    {action_js}'
                         '  }});'
                         '</script>').format(**locals())
        else:
            button_js = ''

        return self.prepare_string(
            '<button type="{button_type}" class="{self.btn_classes}" {data_dismiss} {button_name} id="{button_id}">'
            '{self.label}</button>{button_js}'.format(**locals()), False)


class Actions(object):

    def __init__(self, *args, add_default_crud: bool = False, add_default_filter: bool = False,
                 add_form_buttons: bool = True) -> None:
        super().__init__()
        if len(args) == 1 and args[0] is None:
            self.actions = []
            return

        self.actions = list(args)  # type: List[ActionBase]
        if add_default_crud:
            self.actions.append(
                TableAction(TablePosition.HEADER, _('+ Add'), title=_('Add new record'), name='add',
                            action_js="dynamicforms.newRow('{% url url_reverse|add:'-detail' pk='new' format='html' %}'"
                                      ", 'record', __TABLEID__);")
            )
            self.actions.append(
                TableAction(TablePosition.ROW_CLICK, _('Edit'), title=_('Edit record'), name='edit',
                            action_js="dynamicforms.editRow('{% url url_reverse|add:'-detail' pk='__ROWID__' "
                                      "format='html' %}'.replace('__ROWID__', $(event.target).parents('tr')."
                                      "attr('data-id')), 'record', __TABLEID__);")
            )
            self.actions.append(
                TableAction(TablePosition.ROW_END, label=_('Delete'), title=_('Delete record'), name='delete',
                            action_js="dynamicforms.deleteRow('{% url url_reverse|add:'-detail' pk=row.id %}', "
                                      "{{row.id}}, 'record', __TABLEID__);"))
        if add_default_filter:
            self.actions.append(TableAction(TablePosition.HEADER, label=_('Filter'), title=_('Filter'), name='filter',
                                            action_js="dynamicforms.defaultFilter(event);"))

        if add_form_buttons:
            self.actions.append(FormButtonAction(btn_type=FormButtonTypes.CANCEL, name='cancel'))
            self.actions.append(FormButtonAction(btn_type=FormButtonTypes.SUBMIT, name='submit'))

    def get_resolved_copy(self, serializer) -> 'Actions':
        """
        Returns a copy of declared actions bound to the serializer
        :param serializer: serializer the copy will be bound to
        :return:
        """
        if not isinstance(serializer, Serializer):
            return Actions(None)

        actions = [a.copy_and_resolve_reference(serializer) for a in self.actions]

        # move actions from Field to Serializer
        actions.extend([a.copy_and_resolve_reference(serializer)
                        for field in serializer.fields.values()
                        for a in getattr(field, 'actions', Actions(None)).actions_not_suppressed(serializer)
                        if isinstance(a, FieldChangeAction)])

        return Actions(*actions, add_form_buttons=False)

    def render_field_onchange(self, serializer):
        """
        renders all field onchange actions needed for dynamicforms.registerFieldAction() function

        :return: the actions rendered as template string
        """
        res = ''
        for action in self.actions:
            if isinstance(action, FieldChangeAction) and not isinstance(action, FieldInitAction):
                res += action.render(serializer)
        return res

    def render_form_init(self, serializer):
        """
        renders the function which will analyse initial form data and hide appropriate fields

        :return: the actions rendered as template string
        """
        res = ''
        for action in self.actions:
            if isinstance(action, FormInitAction):
                res += action.render(serializer)
        return res

    def render_field_init(self, serializer, field_name: str):
        """
        renders function that will initialise the field being rendered

        :return: the actions rendered as template string
        """
        res = ''
        for action in self.actions:
            if isinstance(action, FieldInitAction) and field_name in action.tracked_fields:
                res += action.render(serializer)
        return res

    def renderable_actions(self, serializer: Serializer):
        request = serializer.context.get('request', None)
        viewset = serializer.context.get('view', None)
        return (a for a in self.actions
                if isinstance(a, TableAction) and not serializer.suppress_action(a, request, viewset))

    def render_renderable_actions(self, allowed_positions: Iterable[TablePosition], field_name: str,
                                  serializer: Serializer):
        """
        Returns those actions that are not suppressed
        :return: List[Action]
        """
        res = ''

        for action in self.renderable_actions(serializer):
            if action.position in allowed_positions and (field_name is None or field_name == action.field_name):
                res += action.render(serializer)
        return res

    def actions_not_suppressed(self, serializer: Serializer):
        request = serializer.context.get('request', None)
        viewset = serializer.context.get('view', None)

        return (a for a in self.actions if not serializer.suppress_action(a, request, viewset))

    def render_form_buttons(self, serializer: Serializer, position: str):
        """
        Renders form buttons
        :return: List[Action]
        """
        request = serializer.context.get('request', None)
        viewset = serializer.context.get('view', None)

        res = ''

        for button in self.actions:
            if isinstance(button, FormButtonAction) and position in button.positions and \
                    not serializer.suppress_action(button, request, viewset):
                res += button.render(serializer, position=position)
        return res

    def __iter__(self):
        return iter(self.actions)


"""
tipi akcij
    crud (new, edit, delete, view detail)
    on field change
    on form display (initial fields hiding)
    dialog / form manipulation (cancel entry, submit entry)
    custom actions

kako se akcije renderajo
    gumb / link                    -> html control
    onclick / onrightclick / onkey -> JS event
    on field change                -> JS event (depends on field / position)


akcije na formi / dialogu
    control column (start, end of row, any position?)  -> HTML control
    table header / footer          -> html control
    table field left / right       -> html control
    form field left / right        -> html control
    form top, form bottom          -> html control
    custom pozicija  (programer pokliče render funkcijo z nekim custom tekstom, namesto z enum vrednostjo)



    ali morajo vedeti, na čem so?
    pri formi imamo submit, pri dialogu pa dynamicforms.SubmitForm, se pravi, da je render drugačen
    pri tabeli row id, morda row data?


    podatki, ki so lahko na voljo:
    parent serializer
    serializer
    row data
    form data --> serializer
"""



from dynamicforms.action import Actions


class ActionMixin(object):
    """
    Used in fields allowing declaration of actions that happen when field values change
    """

    def __init__(self, *args, actions: Actions = None, **kwargs):
        super().__init__(*args, **kwargs)
        act = actions or Actions(None)
        act.actions.extend(getattr(self, 'actions', Actions()).actions)
        # Obtain a personalised list of actions
        self.actions = act.get_resolved_copy(self)



import re
from datetime import datetime
from typing import Any

from django.utils import timezone
from rest_framework.fields import DateField, TimeField
from rest_framework.serializers import ListSerializer


class NaturalDateTimeMixin(object):
    """
    Used for rendering datetime in human natural style (e.g.: 1 hour, 10 minutes ago)
    """

    def __init__(self, *args, table_format: str = '', **kwargs) -> None:
        """

        :param args:
        :param table_format: Format for datetime rendering in table (non editable). If format is %N:{precision},
            datetime will render in natural style with {precision} depth
        :param kwargs:
        """
        super().__init__(*args, **kwargs)
        self.table_format = table_format

    # noinspection PyUnresolvedReferences
    def to_representation(self, value: Any) -> Any:
        if not value:
            return None

        if isinstance(value, str):
            return value

        if isinstance(self.parent.parent, ListSerializer):
            output_format = getattr(self, 'table_format', None)
            if output_format is not None:
                if re.match(r'%N:\d+', output_format):
                    imported = False
                    try:
                        # This library (will-natural) is used because is the only one we could find that adds text
                        # saying when is this datetime ("ago" or "from now") and have the possibility to set max
                        # precision

                        # noinspection PyPackageRequirements
                        from natural.date import duration
                        imported = True
                    except:
                        print('Install library for natural presentation of date (pip install will-natural)')

                    if imported:
                        if isinstance(self, DateField):
                            now = timezone.now().date()
                        elif isinstance(self, TimeField):
                            now = datetime.now()
                            value = datetime.now().replace(hour=value.hour, minute=value.minute, second=value.second,
                                                           microsecond=value.microsecond)
                        else:
                            now = timezone.now()

                        # noinspection PyUnboundLocalVariable
                        return duration(value, now=now, precision=int(output_format.split(':')[1]))
                else:
                    global_format = getattr(self, 'format', None)
                    setattr(self, 'format', output_format)
                    value = super().to_representation(value)
                    setattr(self, 'format', global_format)
                    return value
        return super().to_representation(value)


class TimeFieldMixin(NaturalDateTimeMixin):

    def __init__(self, *args, table_format: str = None, **kwargs) -> None:
        super().__init__(*args, table_format=table_format, **kwargs)


class DateFieldMixin(NaturalDateTimeMixin):

    def __init__(self, *args, table_format: str = None, **kwargs) -> None:
        super().__init__(*args, table_format=table_format, **kwargs)


class DateTimeFieldMixin(NaturalDateTimeMixin):

    def __init__(self, *args, table_format: str = None, **kwargs) -> None:
        super().__init__(*args, table_format=table_format, **kwargs)



from django.core.cache import cache
from django.http import HttpResponse
from rest_framework.renderers import JSONRenderer


def get_progress_value(request):
    """
    Returns operations progress to HTTP client

    :param request:
    :return: Progress in percentages (0 - 100) if progress is set else None
    """
    progress_key = get_progress_key(request)
    value = None
    if progress_key is not None:
        cache_value = cache.get('df_progress.%s' % progress_key)
        if cache_value is not None:
            value = '%.2f' % cache.get('df_progress.%s' % progress_key)

    return HttpResponse(JSONRenderer().render(
        dict(
            value=value,
            comment=cache.get('df_progress_comment.%s' % progress_key)
        )
    ), content_type='application/json')


def get_progress_key(request):
    """
    Gets progress/operation key under which operation progress will be stored in servers cache.
    :param request:
    :return: Progress key or None if there is no x_df_timestamp in request header
    """
    timestamp = request.META.get('HTTP_X_DF_TIMESTAMP', request.GET.get('x_df_timestamp', None))
    if timestamp is not None:
        return '%s|%s' % (timestamp, request.session.session_key)
    return None


def set_progress_value(progress_key, value):
    """
    Updates operations progress

    :param progress_key: Key under which operations progress is stored
    :param value: Progress in percentages (0.00 - 100.00)
    :return:
    """
    if progress_key is not None:
        cache.set('df_progress.%s' % progress_key, value)


def set_progress_comment(progress_key, value):
    """
    Updates operations comment

    :param progress_key: Key under which operations progress is stored
    :param value: comment
    :return:
    """
    if progress_key is not None:
        cache.set('df_progress_comment.%s' % progress_key, value)



from datetime import datetime, timedelta
from typing import List, Union

import pytz
from django.conf import settings
from django.contrib.auth.views import redirect_to_login
from django.db import models
from django.http import Http404
from rest_framework import status, viewsets
from rest_framework.exceptions import ValidationError
from rest_framework.renderers import JSONRenderer
from rest_framework.response import Response
from rest_framework.serializers import ListSerializer
from rest_framework.utils.serializer_helpers import ReturnDict, ReturnList

from dynamicforms.fields import BooleanField
from .renderers import TemplateHTMLRenderer
from .settings import DYNAMICFORMS


class NewMixin(object):
    """
    Provides support for retrieving default values for a new record.

    Caution: Do not use directly. This is only a mixin and is used in final ViewSet derivatives.
    """

    def new_object(self: viewsets.ModelViewSet):
        """
        Returns a new model instance. If you need it pre-populated with default values, this is the method to override.

        :return: model instance
        """
        # TODO: This function must return an object that has its field values correctly / realistically filled out
        # Example: if a certain field's value hides (sets another field to None), then that second field can't be
        # set to 42, can it?
        # can we run some kind of validation to enforce this?
        # Not that easy: the returned record may not validate for its (correctly) empty fields
        # Maybe we will have to run JavaScript onchange for all fields displayed to ensure at least some consistency?
        # If we do not, subsequent validation may fail because a hidden field has a value
        return self.get_queryset().model()

    # noinspection PyUnresolvedReferences
    def retrieve(self: viewsets.ModelViewSet, request, *args, **kwargs):
        try:
            if not hasattr(super(), 'retrieve'):
                raise Http404()  # This is not a ModelViewSet, so we don't have a retrieval mechanism
            return super().retrieve(request, *args, **kwargs)
        except Http404:
            lookup_url_kwarg = self.lookup_url_kwarg or self.lookup_field
            filter_kwargs = {self.lookup_field: self.kwargs[lookup_url_kwarg]}
            if filter_kwargs.get('pk', None) == 'new':
                instance = self.new_object()
                serializer = self.get_serializer(instance)
                return Response(serializer.data)
            else:
                raise


class PutPostMixin(object):
    """
    Provides support for when there is no record id in URL when calling PUT
    (First empty form is loaded. Than user loads existing data to this form and updates it... - Perform PUT action)

    or

    Provides support for when there is record id in URL when calling POST
    (First form for some existing record is loaded. Than user wants to create new record for it.
    - deletes redord ID and perform POST action)
    """

    # When there is no record id in URL when calling PUT, this function will be called
    # noinspection PyUnresolvedReferences
    def put(self: viewsets.ModelViewSet, request, *args, **kwargs):
        self.kwargs['pk'] = request.data['id']
        return self.update(request, *args, **kwargs)

    # When there is record id in URL when calling POST, this function will be called
    # noinspection PyUnresolvedReferences
    def post(self: viewsets.ModelViewSet, request, *args, **kwargs):
        return self.create(request, *args, **kwargs)


class TemplateRendererMixin():
    template_context = {}
    """
    template_context provides configuration to templates being rendered

    Note that when adding data from database, it is advised to make this definition a callable so that it is evaluated
    on each render of the ViewSet: that way you can ensure data is always loaded fresh from database.

    e.g.

    .. code-block:: python

       template_context = lambda self: dict(items=MyModel.objects.all())

    or

    .. code-block:: python

       def template_context(self):
           return dict(items=MyModel.objects.all())
    """

    template_name = DYNAMICFORMS.table_base_template  #: template filename for listing multiple records (html renderer)

    # noinspection PyAttributeOutsideInit
    def initialize_request(self, request, *args, **kwargs):
        # Caution: just to be sure for any future debugging: the request parameter to this function is a WSGIRequest
        #  while the return Request is actually DRF Request
        #  As a consequence, form values don't get parsed until you actually call super().initialize_request
        #  There's no "request.data", etc. Just saying. So you don't debug for two hours next time. By "you" I mean me

        # Force render using a given render path (full page, table, table rows, form, dialog with form)
        self.render_type = request.META.get('HTTP_X_DF_RENDER_TYPE', request.GET.get('df_render_type', 'page'))

        if request.method.lower() == 'post' and request.POST.get('data-dynamicforms-method', None):
            # This is a hack because HTML forms can only do POST & GET. This way we also get PUT & PATCH
            request.method = request.POST.get('data-dynamicforms-method')
            # If we don't set this META, django won't recognise our CSRF token
            request.META['HTTP_X_CSRFTOKEN'] = request.POST['csrfmiddlewaretoken']
        return super().initialize_request(request, *args, **kwargs)

    def finalize_response(self, request, response, *args, **kwargs):
        if not isinstance(response, Response):
            return response

        res = super().finalize_response(request, response, *args, **kwargs)

        def get_query_params():
            if request.query_params:
                return '?' + '&'.join(['%s=%s' % (key, value) for key, value in request.query_params.items()])
            return ''

        if isinstance(res.accepted_renderer, TemplateHTMLRenderer):
            if status.is_success(res.status_code) or res.status_code == status.HTTP_400_BAD_REQUEST:
                if isinstance(res.data, dict) and 'next' in res.data and 'results' in res.data and \
                        isinstance(res.data['results'], (ReturnList, ReturnDict)):
                    serializer = res.data['results'].serializer
                else:
                    serializer = res.data.serializer

                if isinstance(serializer, ListSerializer):
                    serializer.child.render_type = self.render_type
                else:
                    serializer.render_type = self.render_type

                if self.render_type in ('table', 'table rows'):
                    serializer.data_template = self.template_name
                elif self.render_type == 'dialog':
                    serializer.data_template = DYNAMICFORMS.modal_dialog_template
                    res.template_name = DYNAMICFORMS.modal_dialog_template
                elif self.render_type == 'form':
                    serializer.data_template = res.data.serializer.template_name
                    res.template_name = res.data.serializer.template_name
                else:
                    if isinstance(serializer, ListSerializer):
                        serializer.child.render_type = 'table'
                        serializer.child.data_template = self.template_name
                    else:
                        serializer.render_type = 'form'
                        serializer.data_template = serializer.template_name
            elif res.status_code in (status.HTTP_401_UNAUTHORIZED, status.HTTP_403_FORBIDDEN) and \
                    self.render_type != 'dialog':
                # TODO: We should show a message here that user is not authorized for this action (only for 403)
                res = redirect_to_login(request.path_info + get_query_params())
        return res


class ModelViewSet(NewMixin, PutPostMixin, TemplateRendererMixin, viewsets.ModelViewSet):
    """
    In addition to all the functionality, provided by DRF, DynamicForms ViewSet has some extra features:

    * Separate templates for rendering list or single record
    * You can request a "new" record and even have it pre-populated with values
    * To render viewset as API or JSON use the same method as in DRF: To render it in HTML just add ".html" to the URL.
    * Standard DRF router URL patterns apply:

       * To render a new record use pk=new.
       * To render an existing record (for editing) use pk={record_id}.

    """

    def paginate_queryset(self, queryset):
        """
        Return a single page of results, or `None` if pagination is disabled.
        """
        if self.paginator is None:
            return None
        # determine request format and handle pagination for json format
        if isinstance(self.request.accepted_renderer, JSONRenderer) and not BooleanField().to_internal_value(
                self.request.META.get('HTTP_X_PAGINATION', self.request.GET.get('x_df_pagination', False))):
            return None
        return self.paginator.paginate_queryset(queryset, self.request, view=self)

    def get_queryset(self):
        """
        Returns records from queryset with filters applied

        :return: filtered records
        """
        return super().get_queryset()

    def filter_queryset(self, queryset, query_params=None):
        """
        Applies filters for all fields

        :param queryset: Queryset
        :param query_params: Custom query_params if needed
        :return: queryset with filters applied
        """
        res = queryset
        if self.request:
            if query_params is None:
                query_params = self.request.query_params
            for fld, val in query_params.items():
                res = self.filter_queryset_field(res, fld, val)
        return res

    # noinspection PyMethodMayBeStatic
    def filter_queryset_field(self, queryset, field, value):
        """
        Applies filter to individual field

        :param queryset: Queryset
        :param field: Field name
        :param value: Field value
        :return: queryset with applied filter for the field
        """
        if value is None or value == '':
            return queryset

        model_meta = queryset.model._meta

        if field not in (fld.name for fld in model_meta.get_fields()):
            return queryset

        # TODO: this would probably be better moved into the fields themselves
        if isinstance(model_meta.get_field(field), (models.CharField, models.TextField)):
            return queryset.filter(**{field + '__icontains': value})
        if isinstance(model_meta.get_field(field), (models.DateField, models.DateTimeField)):
            date_time = None
            for date_time_fmt in [settings.DATETIME_FORMAT, '%Y-%m-%dT%H:%M:%S', settings.DATE_FORMAT, '%Y-%m-%d']:
                try:
                    date_time = datetime.strptime(value, date_time_fmt)
                    break
                except:
                    pass
            if date_time is None:
                return queryset
            date_time = pytz.timezone(settings.TIME_ZONE).localize(date_time).astimezone(pytz.utc)
            if len(value) <= 10:
                return queryset.filter(**{field + '__gte': date_time, field + '__lt': date_time + timedelta(days=1)})
            return queryset.filter(**{field + '__gte': date_time, field + '__lt': date_time + timedelta(seconds=1)})
        else:
            if isinstance(model_meta.get_field(field), models.BooleanField):
                value = (value == 'true')
            return queryset.filter(**{field: value})

    @staticmethod
    def generate_paged_loader(page_size: int = 30, ordering: Union[str, List[str]] = 'id'):
        """
        Generates a Pagination class that will handle dynamic data loading for ViewSets with a lot of data.
        Use by declaring `pagination_class = ModelViewSet.generate_paged_loader()` in class variables

        :param page_size: how many records should be fetched at a time
        :param ordering: This should be a string, or list of strings, indicating the field against which the cursor
           based pagination will be applied. For example: ordering = 'slug'
        :return: a Pagination class
        """
        from rest_framework.pagination import CursorPagination
        ps = page_size
        ordr = ordering

        class MyCursorPagination(CursorPagination):
            ordering = ordr
            page_size = ps
            df_request = None

            def paginate_queryset(self, queryset, request, view=None):
                self.df_request = request
                return super().paginate_queryset(queryset, request, view=None)

            def encode_cursor(self, cursor):
                # Following code is needed when we have https proxy server that redirects requests to http servers.
                # In that case original code generates cursor links that have http scheme.
                # So here I check REFERER header to find out which scheme is originally declared.
                # And use that one in cursor link.
                request = getattr(self, 'df_request', None)
                cursor_url = super().encode_cursor(cursor).split(':', 1)
                req_url = self.df_request.META.get('HTTP_REFERER', None)
                if req_url:
                    req_url = req_url.split(':', 1)
                    if cursor_url[0] != req_url[0] and req_url[0].lower() in ('http', 'https'):
                        cursor_url[0] = req_url[0]
                cursor_url = ':'.join(cursor_url)
                if request and isinstance(request.accepted_renderer, JSONRenderer):
                    cursor_url += '&x_df_pagination=1'
                return cursor_url

        return MyCursorPagination

    def handle_create_validation_exception(self, e, request, *args, **kwargs):
        instance = self.new_object()
        ser = self.get_serializer(instance, data=request.data, partial=False)
        ser.is_valid(raise_exception=False)
        e.detail.serializer = ser
        raise e

    def create(self, request, *args, **kwargs):
        try:
            return super().create(request, *args, **kwargs)
        except ValidationError as e:
            self.handle_create_validation_exception(e, request, *args, **kwargs)


class SingleRecordViewSet(NewMixin, TemplateRendererMixin, viewsets.GenericViewSet):

    def new_object(self):
        raise NotImplementedError()

    def create(self, request, *args, **kwargs):
        raise NotImplementedError()


# noinspection PyUnresolvedReferences
class GenericViewSet(NewMixin, PutPostMixin, TemplateRendererMixin, viewsets.GenericViewSet):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



from typing import Optional


class RelatedFieldAJAXMixin(object):

    def __init__(self, *args, url_reverse: Optional[str] = None, placeholder: Optional[str] = None,
                 additional_parameters: Optional[dict] = None, query_field: str = 'query', **kwargs):
        """
        Allows us to use AJAX to populate select2 options instead of pre-populating at render time

        :param args:
        :param url_reverse: reverse url to ViewSet providing the JSON data
        :param placeholder: select2 placeholder to display until user selects a value
        :param additional_parameters: additional parameters to be sent to ViewSet as part of the query
        :param query_field: field against which user search will be performed
        :param kwargs:
        """
        super().__init__(*args, **kwargs)
        self.url_reverse = url_reverse
        self.placeholder = placeholder
        self.additional_parameters = additional_parameters
        self.query_field = query_field

    @property
    def additional_parameters_urlencoded(self):
        from django.utils.http import urlencode
        return '?' + urlencode(self.additional_parameters)

    # noinspection PyUnresolvedReferences
    def iter_options_bound(self, value):
        if self.url_reverse:
            qry = self.get_queryset()
            try:
                qry = qry.filter(pk=value)
                return [dict(value=value, display_text=self.display_value(qry.first()))]
            except:
                return []
        return super().iter_options()



__title__ = 'DynamicForms'
__version__ = '0.9.9'
__author__ = 'Jure Erznožnik'
__email__ = 'jure.erznoznik@gmail.com'
__license__ = 'BSD 3-Clause'
__copyright__ = 'Copyright 2017-present Jure Erznožnik'

VERSION = __version__



import uuid as uuid_module
from enum import IntEnum
from typing import Hashable

from rest_framework.fields import Field as DrfField
from rest_framework.relations import ManyRelatedField, RelatedField
from rest_framework.templatetags import rest_framework as drftt

from dynamicforms.settings import DYNAMICFORMS


class DisplayMode(IntEnum):
    SUPPRESS = 1  # Field will be entirely suppressed. it will not render (not even to JSON) and will not parse for PUT
    HIDDEN = 5  # Field will render as <input type="hidden"> or <tr data-field_name>
    INVISIBLE = 8  # Field will render completely, but with display: none. Equal to setting its style = {display: none}
    FULL = 10  # Field will render completely


class RenderMixin(object):
    """
    Is used in fields and serializers, so every field and serializer gets its unique id. Also to specify where and how
    fields should render.

    In form where serializer is used, id is serializers uuid. Table with list of records has id »list-serializer.uuid«,
    in dialog id is »dialog-{serializer.uuid}« and save button's id on dialog is »save-{serializer.uuid}«

    Similar for fields: All inputs in HTML get id from field.uuid. Div that contains all that belongs to the field has
    »container-{field.uuid}« for id, label has »label-{field.uuid}« and help text (if exists) has »help-{field.uuid}«
    for id.

    Used for rendering individual field to table view
    """

    def __init__(self, *args, uuid: uuid_module.UUID = None,
                 display: DisplayMode = None,  # None == Leave at default
                 display_table: DisplayMode = None,  # None == Leave at default
                 display_form: DisplayMode = None,  # None == Leave at default
                 table_classes: str = '',
                 **kwargs):
        """

        :param args: passed on to inherited constructors
        :param uuid: custom specified field UUID. if not specified, it will be assigned automatically
        :param display: see DisplayMode enum. Specifies how field will render. Leave at None for default (FULL)
            display_form and display_table also accepted for better granularity
        :param table_classes: css classes to add to the table column
        :param kwargs: passed on to inherited constructors
        """
        super().__init__(*args, **kwargs)
        self.uuid = uuid or uuid_module.uuid1()
        # noinspection PyUnresolvedReferences
        self.display_table = (
            display_table or display
            or (DisplayMode.FULL if not getattr(self, 'write_only', False) else DisplayMode.SUPPRESS)
        )
        self.display_form = display_form or display or DisplayMode.FULL
        self.table_classes = table_classes

    @property
    def is_rendering_to_list(self):
        """
        reports whether we are currently rendering to table or to single record
        :return:
        """
        try:
            # noinspection PyUnresolvedReferences
            if self.parent.parent:
                return True
        except:
            pass
        return False

    @property
    def is_rendering_to_html(self):
        try:
            # noinspection PyUnresolvedReferences
            return self.context['format'] == 'html'
        except:
            pass
        return False

    # noinspection PyUnresolvedReferences
    def use_pk_only_optimization(self):
        """
        Overrides DRF RelatedField's method. It True is returned then value passed for serailization will be PK value
        only, not entire relation object
        :return:
        """
        if self.is_rendering_to_list and self.is_rendering_to_html:
            return False
        return super().use_pk_only_optimization()

    # noinspection PyUnresolvedReferences
    def to_representation(self, value):
        """
        Overrides DRF Field's to_representation.
        Note that this is also called for the entire record as well as the serializer also is a Field descendant
        :param value: value to serialize
        :return: serialized value
        """
        if self.is_rendering_to_list and self.is_rendering_to_html:
            return self.render_to_table(value, self.parent.instance)
        return super().to_representation(value)

    def set_display(self, value):
        self.display_form = self.display_table = value

    display = property(lambda self: self.display_form, set_display)

    # noinspection PyUnusedLocal, PyUnresolvedReferences
    def render_to_table(self, value, row_data):
        """
        Renders field value for table view

        :param value: field value
        :param row_data: data for entire row (for more complex renderers)
        :return: rendered value for table view
        """
        get_queryset = getattr(self, 'get_queryset', None)
        if isinstance(self, RelatedField) or get_queryset:
            return self.display_value(value)
        elif isinstance(self, ManyRelatedField):
            # if value is a list, we're dealing with ManyRelatedField, so let's not do that
            print('WARNING/TODO: ManyRelatedField lookup wasn\'t fixed with the rest of the code for lack of examples')
            cr = self.child_relation
            return ', '.join((cr.display_value(item) for item in cr.get_queryset().filter(pk__in=value)))
        else:
            choices = getattr(self, 'choices', {})

        # Now that we got our choices for related & choice fields, let's first get the value as it would be by DRF
        value = super().to_representation(value)

        if isinstance(value, Hashable) and value in choices:
            # choice field: let's render display names, not values
            value = choices[value]
        if value is None:
            return DYNAMICFORMS.null_text_table

        return drftt.format_value(value)

    def validate_empty_values(self: DrfField, data):
        # noinspection PyUnresolvedReferences
        res = super().validate_empty_values(data)

        # This is to fix a problem with calculated fields which was only solved in DRF 3.10.
        # Forces validation and inclusion of the field into validated data. See comment in original function.
        if res == (True, None) and data is None and self.source == '*':
            return False, None
        return res



from rest_framework.fields import ChoiceField


class NullChoiceMixin(object):
    """
    Declaring ChoiceField's null_choice_text
      (what should be rendered for "no selection"; DRF's default is ---------)
      We also support user declaring a choice with None value and a different text (see hidden fields example)
    """

    @property
    def null_choice_text(self):
        res = '--------'
        if isinstance(self, ChoiceField) and not getattr(self, 'get_queryset', None):
            # Only do this for true ChoiceFields. ReladedFields would run a query here
            # Possibly we will at some point have to enable this for relatedfields too
            for opt in self.iter_options():
                if opt.start_option_group or opt.end_option_group:
                    pass
                elif opt.value is None:
                    res = opt.display_text
        return res

    def iter_options_bound(self, value):
        # noinspection PyUnresolvedReferences
        return super().iter_options()



class Struct(object):
    def __init__(self, data=None, **kwds):
        if not data:
            data = {}
        for name, value in data.items():
            if name:
                setattr(self, name, self._wrap(value))
        for name, value in kwds.items():
            if name:
                setattr(self, name, self._wrap(value))

    def _wrap(self, value):
        if isinstance(value, (tuple, list, set, frozenset)):
            return type(value)([self._wrap(v) for v in value])
        else:
            return Struct(value) if isinstance(value, dict) else value

    def clone(self, **kwds):
        return Struct(self.__to_dict__(), **kwds)

    def __repr__(self):
        return "Struct: " + repr(self.__dict__)

    def __len__(self):
        return len(self.__dict__)

    def __to_dict__(self):
        res = {}
        res.update(self.__dict__)
        for k in res.keys():
            if isinstance(res[k], Struct):
                res[k] = res[k].__to_dict__()
            elif isinstance(res[k], (tuple, list, set, frozenset)):
                res[k] = [i.__to_dict__() if isinstance(i, Struct) else i for i in res[k]]
        return res


class StructDefault(Struct):
    def __getattr__(self, item):
        return self._default_



from django.http import HttpRequest

from .settings import DYNAMICFORMS


def add_dynamicforms_settings(request: HttpRequest):
    """
    context processor that adds DynamicForms configuration variables to template context

    Example for supporting different versions of bootstrap:

    .. code-block:: python

       {% if DYNAMICFORMS.bootstrap_version == 'v3' %}
         {% set_var card_class='panel panel-default' card_header='panel-heading' card_body='panel_body' %}
       {% else %}
         {% set_var card_class='card' card_header='card-header' card_body='card-body' %}
       {% endif %}

    .. note:: Using DynamicForms renderers automatically adds this variable into the context as it is required by
       template packs.

    :param request: see `django documentation <https://docs.djangoproject.com/en/dev/ref/templates/api/
       #writing-your-own-context-processors>`_
    :return: dict with `DYNAMICFORMS` context variable set
    """
    return dict(DYNAMICFORMS=DYNAMICFORMS)



# flake8: noqa

from .action import ActionMixin
from .help_text import FieldHelpTextMixin
from .implicit_hidden import HiddenFieldMixin
from .implicit_natural_date import DateFieldMixin, DateTimeFieldMixin, TimeFieldMixin
from .null_choice import NullChoiceMixin
from .related_field_ajax import RelatedFieldAJAXMixin
from .render import DisplayMode, RenderMixin



from typing import Optional, Union


class FieldHelpTextMixin(object):
    """
    Redefines help_text property such that we can set help text for api docs and form help text
    """

    def __init__(self, *args, help_text: Union[str, dict, None] = None, help_text_form: Optional[str] = None,
                 **kwargs) -> None:
        super().__init__(*args, help_text=help_text, **kwargs)
        self.help_text = help_text
        if help_text_form or not isinstance(help_text, dict):
            self._help_text_form = help_text_form or self.help_text

    def get_help_text(self):
        return self._help_text

    def set_help_text(self, value):
        if isinstance(value, dict):
            self._help_text = value.get('doc', None)
            self._help_text_form = value.get('form', None)
        else:
            self._help_text = value
            self._help_text_form = getattr(self, '_help_text_form', None) or value

    def get_help_text_form(self):
        return self._help_text_form

    def set_help_text_form(self, value):
        if isinstance(value, dict):
            self._help_text = value.get('form', None)
        else:
            self._help_text_form = value

    help_text = property(get_help_text, set_help_text)
    help_text_form = property(get_help_text_form, set_help_text_form)



from typing import Any

from django.db import models
from rest_framework import serializers

from dynamicforms.action import Actions
from dynamicforms.settings import DYNAMICFORMS
from . import fields
from .mixins import ActionMixin, DisplayMode, RenderMixin
from .struct import StructDefault


class DynamicFormsSerializer(RenderMixin, ActionMixin):
    template_name = DYNAMICFORMS.form_base_template  #: template filename for single record view (HTMLFormRenderer)
    actions = Actions(add_default_crud=True, add_form_buttons=True)
    form_titles = {
        'table': '',
        'new': '',
        'edit': '',
    }

    show_filter = False  # When true, filter row is shown for list view

    def __init__(self, *args, is_filter: bool = False, **kwds):
        self.master = None
        self.is_filter = is_filter
        if self.is_filter:
            try:
                instance = self.Meta.model()
                for fld in instance._meta.fields:
                    setattr(instance, fld.name, None)
            except:
                instance = StructDefault(_default_=None)
            kwds.setdefault('instance', instance)
        super().__init__(*args, **kwds)
        try:
            # hide the primary key field (DRF only marks it as R/O)
            field_name = self.Meta.model._meta.pk.name
            if field_name not in self._declared_fields:
                pk_field = self.fields[field_name]
                pk_field.display_form = fields.DisplayMode.HIDDEN
                pk_field.display_table = fields.DisplayMode.FULL
        except:
            pass
        if self.is_filter:
            for field in self.fields.values():
                field.default = None
                field.allow_blank = True
                field.allow_null = True
                field.read_only = False
                field.display_form = field.display_table  # filter's form is same as non-filter's table

    @property
    def has_non_field_errors(self):
        """
        Reports whether validation turned up any form-wide validation errors. Used in templates to render the form-wide
        error message

        :return: True | False depending on whether form validation failed
        """
        if hasattr(self, '_errors'):
            return 'non_field_errors' in self.errors
        return False

    @property
    def page_title(self):
        """
        Returns page title from form_titles based on the rendered data
        :return string: page title
        """
        if self.render_type == 'table':
            return self.form_titles.get('table', '')
        elif self.data.get('id', None):
            return self.form_titles.get('edit', '')
        else:
            return self.form_titles.get('new', '')

    # noinspection PyProtectedMember
    @property
    def filter_data(self):
        """
        Returns serializer for filter row in table
        :return:  Serializer
        """
        if getattr(self, '_filter_ser', None) is None:
            # noinspection PyAttributeOutsideInit
            self._filter_ser = type(self)(is_filter=True, context=getattr(self, 'context', {}))
            self._filter_ser.master = self
        return self._filter_ser  # Just create the same serializer in filter mode (None values, allow_nulls)

    # noinspection PyUnusedLocal
    def suppress_action(self, action, request, viewset):
        """
        Determines whether rendering an action into the DOM should be suppressed. Use when some users don't have access
        to some of the functionality, e.g. when CRUD functionality is only enabled for administrative users
        :param action: action to be checked
        :param request: request that triggered the render (may be None)
        :param viewset: viewset that provided the serialized data (may be None)
        :return: boolean whether action should render (False) or not (True)
        """
        return False

    @property
    def renderable_actions(self: 'serializers.Serializer'):
        """
        Returns those actions that are not suppressed
        :return: List[Action]
        """
        request = self.context.get('request', None)
        viewset = self.context.get('view', None)
        return [action for action in self.controls.actions if not self.suppress_action(action, request, viewset)]

    # noinspection PyUnresolvedReferences
    def get_initial(self) -> Any:
        if getattr(self, '_errors', None):
            # This basically reproduces BaseSerializer.data property except that it disregards the _errors member
            if self.instance:
                res = self.to_representation(self.instance)
            # elif hasattr(self, '_validated_data'):
            #     res = self.to_representation(self.validated_data)
            else:
                res = {}
            res.update(super().get_initial())

            return res

        return super().get_initial()


class ModelSerializer(DynamicFormsSerializer, serializers.ModelSerializer):
    """
    DynamicForms' ModelSerializer overrides the following behaviour over DRF's implementation:

    * Uses own field types for construction
    * Adds form UUID (rendered in html too)
    * Adds processing for form-wide errors

    DRF's docstring copied verbatim:

    A `ModelSerializer` is just a regular `Serializer`, except that:

    * A set of default fields are automatically populated.
    * A set of default validators are automatically populated.
    * Default `.create()` and `.update()` implementations are provided.

    The process of automatically determining a set of serializer fields
    based on the model fields is reasonably complex, but you almost certainly
    don't need to dig into the implementation.

    If the `ModelSerializer` class *doesn't* generate the set of fields that
    you need you should either declare the extra/differing fields explicitly on
    the serializer class, or simply use a `Serializer` class.
    """

    def __init__(self, *args, is_filter: bool = False, **kwds):
        if hasattr(self.Meta, 'fields'):
            if self.Meta.fields != '__all__' and 'df_prev_id' not in self.Meta.fields:
                self.Meta.fields += 'df_prev_id',
        super().__init__(*args, is_filter=is_filter, **kwds)
        self.manage_changed_flds()

    serializer_field_mapping = {
        models.AutoField: fields.IntegerField,
        models.BigIntegerField: fields.IntegerField,
        models.BooleanField: fields.BooleanField,
        models.CharField: fields.CharField,
        models.CommaSeparatedIntegerField: fields.CharField,
        models.DateField: fields.DateField,
        models.DateTimeField: fields.DateTimeField,
        models.DecimalField: fields.DecimalField,
        models.EmailField: fields.EmailField,
        models.Field: fields.ModelField,
        models.FileField: fields.FileField,
        models.FloatField: fields.FloatField,
        models.ImageField: fields.ImageField,
        models.IntegerField: fields.IntegerField,
        models.NullBooleanField: fields.NullBooleanField,
        models.PositiveIntegerField: fields.IntegerField,
        models.PositiveSmallIntegerField: fields.IntegerField,
        models.SlugField: fields.SlugField,
        models.SmallIntegerField: fields.IntegerField,
        models.TextField: fields.CharField,
        models.TimeField: fields.TimeField,
        models.URLField: fields.URLField,
        models.GenericIPAddressField: fields.IPAddressField,
        models.FilePathField: fields.FilePathField,
    }
    if models.DurationField is not None:
        serializer_field_mapping[models.DurationField] = fields.DurationField

    serializer_related_field = fields.PrimaryKeyRelatedField
    serializer_related_to_field = fields.SlugRelatedField
    serializer_url_field = fields.HyperlinkedIdentityField
    serializer_choice_field = fields.ChoiceField

    def manage_changed_flds(self):
        """
        When there is a need to only change few parameters of a field put those fields and changed parameters in
        serializers Meta class in parameter changed_flds.

        Exsample:
        changed_flds = {'id': dict(display=DisplayMode.HIDDEN),
                        'comment': dict(label='Comm', help_text='Help text for comment field')
                        }
        :return:
        """
        if hasattr(self.Meta, 'changed_flds'):
            for field, params in self.Meta.changed_flds.items():
                field_def = self.fields.get(field, None)
                if field_def:
                    for key, val in params.items():
                        setattr(field_def, key, val)

    # Dynamic forms default field that is used to contain data for positioning (id of previous record)
    df_prev_id = fields.SerializerMethodField(display=DisplayMode.HIDDEN)

    def fetch_prev_id(self, obj, view):
        ordering = 'id'
        try:
            ordering = view.pagination_class.ordering
        except:
            pass
        if ordering != 'id':
            query_params = self.context['request'].query_params
            query_params._mutable = True
            query_params.pop('id', '')
            queryset = view.filter_queryset(queryset=view.get_queryset(), query_params=query_params)
            records = list(queryset.order_by(ordering))
            curr_index = records.index(obj)
            prev_id = None
            if curr_index > 0:
                prev_id = records[curr_index - 1].id

            return prev_id

        return ''

    # noinspection PyMethodMayBeStatic
    def get_df_prev_id(self, obj):
        try:
            if self.context['request'].META.get('HTTP_X_DF_CALLTYPE', '') == 'refresh_record':
                view = self.context.get('view', None)
                if view:
                    return self.fetch_prev_id(obj, view)
        except:
            pass

        return ''


class Serializer(DynamicFormsSerializer, serializers.Serializer):

    def update(self, instance, validated_data):
        # Implemented just so IDE doesn't complain. Normally this will be handled in SingleRecordViewSet
        pass

    def create(self, validated_data):
        # Implemented just so IDE doesn't complain. Normally this will be handled in SingleRecordViewSet
        pass



from .render import DisplayMode, RenderMixin


class HiddenFieldMixin(RenderMixin):

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('display_table', DisplayMode.SUPPRESS)
        super().__init__(*args, **kwargs)



from django.conf import settings as s

from .struct import Struct

DYNAMICFORMS_BOOTSTRAP = 'dynamicforms/bootstrap/'
DYNAMICFORMS_JQUERY_UI = 'dynamicforms/jquery_ui/'


class Settings(Struct):

    # Support for jQueryUI (this will be queried in dynamicforms.js)
    jquery_ui = False

    # if True, base_list template will render the HTML such that it will display dialog
    #   Make sure you {% include DYNAMICFORMS.modal_dialog_template %} somewhere top-level in your html body
    # if False, the record will be displayed in a new page
    edit_in_dialog = True

    # Support for select2
    use_select2 = True

    # specifies the basepage template to be used for TemplateHTMLRenderer
    page_template = ''

    # specifies text to be displayed in grid when field's value is null
    null_text_table = 'null'

    def __init__(self, data=None, **kwds):
        kwds.setdefault('page_template', DYNAMICFORMS_BOOTSTRAP + 'page.html')
        super().__init__(data, **kwds)
        self.template = DYNAMICFORMS_BOOTSTRAP

    def __to_dict__(self):
        res = {}
        res.update(super().__to_dict__())
        for p in dir(self):
            if not (p == 'clone' or p.startswith('_')):
                res[p] = getattr(self, p)
        return res

    # noinspection PyPep8Naming
    @property
    def DisplayMode(self):
        from .mixins import DisplayMode
        return {e.name: e.value for e in DisplayMode}  # A copy to be accessible in the templates

    # ****************************************************************************
    # These are constants, generated from settings. They are shortcuts for quick use in templates and are specific
    # to the chosen template pack.
    # ****************************************************************************

    select2_form_field_class = property(lambda self: 'select2-field' if self.use_select2 else '')

    # Path to template for :samp:`<html><head>` tag JS & CSS includes, necessary for the chosen template pack.
    page_includes = property(lambda self: self.template + 'base_includes.html')

    # Path to base template for fields.
    field_base_template = property(lambda self: self.template + 'field/base_field.html')

    # Path to base template for forms.
    form_base_template = property(lambda self: self.template + 'base_form.html')

    # Path to base template for forms.
    table_base_template = property(lambda self: self.template + 'base_list.html')

    # Path to base template for forms.
    table_filter_base_template = property(lambda self: self.template + 'base_table_filter.html')

    # Path to base template for table body.
    table_body_base_template = property(lambda self: self.template + 'base_table_body.html')

    # Path to template for modal dialog
    modal_dialog_template = property(lambda self: self.template + 'modal_dialog.html')

    # classes to use on form buttons
    form_button_classes = property(lambda self: '')
    form_button_classes_cancel = property(lambda self: '')
    form_button_classes_primary = property(lambda self: '')
    form_button_classes_secondary = property(lambda self: '')


def if3_4(if3, if4):
    def inner(self):
        if self.bootstrap_version == 'v4':
            return if4
        return if3

    return inner


class SettingsBootstrap(Settings):
    # Supported Bootstrap versions are v3 and v4
    bootstrap_version = 'v4'

    def __init__(self, data=None, **kwds):
        super().__init__(data, **kwds)
        self.template = DYNAMICFORMS_BOOTSTRAP

    # select2 helper variables
    select2_theme = property(if3_4('bootstrap', 'bootstrap4'))
    select2_include = property(lambda self: self.template + 'base_includes_select2.html')
    select2_bs_css = property(if3_4('select2/css/select2-bootstrap3.min.css', 'select2/css/select2-bootstrap4.min.css'))

    # classes for card divs
    bs_card_class = property(if3_4('panel panel-default', 'card'))
    bs_card_header = property(if3_4('panel-heading df-card-header', 'card-header df-card-header'))
    bs_card_body = property(if3_4('panel-body df-card-body', 'card-body df-card-body'))

    page_includes = property(lambda self: self.template + ('base_includes_%s.html' % self.bootstrap_version))
    field_base_template = property(lambda self: self.template + ('field/base_field_%s.html' % self.bootstrap_version))
    modal_dialog_template = property(lambda self: self.template + ('modal_dialog_%s.html' % self.bootstrap_version))

    # classes to use on form buttons
    form_button_classes = property(lambda self: 'btn ml-1')
    form_button_classes_cancel = property(lambda self: '')
    form_button_classes_primary = property(lambda self: 'btn-primary')
    form_button_classes_secondary = property(lambda self: 'btn-secondary')


class SettingsJqueryUI(Settings):
    jquery_ui = True

    def __init__(self, data=None, **kwds):
        kwds.setdefault('page_template', DYNAMICFORMS_JQUERY_UI + 'page.html')
        super().__init__(data, **kwds)
        self.template = DYNAMICFORMS_JQUERY_UI

    # classes to use on form buttons
    form_button_classes = property(lambda self: 'ui-button ui-corner-all ui-widget')
    form_button_classes_cancel = property(lambda self: 'close-btn')
    form_button_classes_primary = property(lambda self: 'ui-state-highlight')
    form_button_classes_secondary = property(lambda self: '')


def get_settings():
    """
    Let's parse settings from settings.py

    :return: Settings class
    """
    df = Struct(
        template=DYNAMICFORMS_BOOTSTRAP,
    )
    df = df.clone(**getattr(s, 'DYNAMICFORMS', {}))
    template = df.template
    if template == DYNAMICFORMS_BOOTSTRAP:
        return SettingsBootstrap(**df.__to_dict__())
    return SettingsJqueryUI(**df.__to_dict__())


DYNAMICFORMS = get_settings()


def version_check(checked_version, min_version):
    """
    Checks whether checked version is higher or equal to given min version
    :param checked_version: version being checked
    :param min_version: minimum allowed version
    :return: True when checked version is high enough
    """

    def version_transform(ver):
        version = ''
        vn = va = ''
        stage = 0
        ver = '.' + (ver or '')
        for c in ver:
            if c == '.':
                if vn or va:
                    version += '{0:0>3}{1: <2}.'.format(vn, va)
                vn = va = ''
                stage = 0
                continue
            if c.isdigit():
                pass
            elif c.isalpha():
                stage = max(1, stage)
            else:
                stage = max(2, stage)
            if stage == 0:
                vn += c
            elif stage == 1:
                va += c
        if vn or va:
            version += '{0:0>3}{1: <2}.'.format(vn, va)
        return version[:-1]

    if not checked_version:
        return False
    return version_transform(checked_version) >= version_transform(min_version)



from django.conf.urls import url

from dynamicforms import progress

urlpatterns = [
    # Progress is used for checking on progress of operation on server
    url(r'^progress/$', progress.get_progress_value, name='progress'),
]



import random

import django.utils.timezone

from examples.models import Filter, PageLoad, Relation


def add_relation(apps, schema_editor):
    if not Relation.objects.all().count():
        for i in range(1, 11):
            Relation.objects.create(name='Relation object %d' % i)


def add_filter(apps, schema_editor):
    if not Filter.objects.all().count():
        date_now = django.utils.timezone.now()
        char_field = ['abc', 'def', 'ghi', 'abcdef', 'abcghi', 'defghi', 'abcdefghi']
        for i in range(1, 100 + 1):
            date_now = date_now + django.utils.timezone.timedelta(hours=3)
            Filter.objects.create(char_field='%s %d' % (char_field[(i - 1) % 7], i),
                                  datetime_field=date_now,
                                  int_field=((i - 1) % 10) + 1,
                                  int_choice_field=((i - 1) % 4),
                                  bool_field=((i - 1) % 10) > 4
                                  )


def add_page_load(apps, schema_editor):
    if not PageLoad.objects.all().count():
        for i in range(1, 10000 + 1):
            PageLoad.objects.create(description='Item %d' % i, choice=random.randint(1, 4))


class DynamicformsMigrationMixin(object):
    def apply(self, project_state, schema_editor, collect_sql=False):
        applied = super().apply(project_state, schema_editor, collect_sql)
        if '0003_filter_name' in self.name:
            add_relation(None, None)
            add_page_load(None, None)
            add_filter(None, None)
        return applied



import setuptools

with open('README.rst', 'r') as fh:
    long_description = fh.read()
with open('requirements.txt', 'r') as fh:
    requirements = fh.readlines()

setuptools.setup(
    name="DynamicForms",
    version="0.9.9",
    author="Jure Erznožnik",
    author_email="jure@velis.si",
    description="DynamicForms performs all the visualisation & data entry of your DRF Serializers & ViewSets and adds "
                "some candy of its own: It is a django library that gives you the power of dynamically-shown form "
                "fields, auto-filled default values, dynamic record loading and similar candy with little effort. "
                "To put it differently: once defined, a particular ViewSet / Serializer can be rendered in multiple "
                "ways allowing you to perform viewing and authoring operations on the data in question.",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/velis74/DynamicForms",
    packages=setuptools.find_packages(include=('dynamicforms',)),
    include_package_data=True,
    install_requires=requirements,
    python_requires='>=3.4',
    license='BSD-3-Clause',
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Framework :: Django",
    ],
)



import copy

import six
from django.template import loader
from rest_framework.renderers import HTMLFormRenderer, TemplateHTMLRenderer
from rest_framework.serializers import HiddenField, ListSerializer
from rest_framework.utils.serializer_helpers import ReturnDict, ReturnList

from dynamicforms.fields import BooleanField
from .settings import DYNAMICFORMS


# noinspection PyRedeclaration
class TemplateHTMLRenderer(TemplateHTMLRenderer):
    """
    Renderers serializer data into an HTML form.

    If the serializer was instantiated without an object then this will
    return an HTML form not bound to any object,
    otherwise it will return an HTML form with the appropriate initial data
    populated from the object.

    Note that rendering of field and form errors is not currently supported.
    """

    def render(self, data, accepted_media_type=None, renderer_context=None):
        link_next = link_prev = ''

        if isinstance(data, dict) and 'next' in data and 'results' in data and \
                isinstance(data['results'], (ReturnList, ReturnDict)):
            # This is in case of Pagination
            link_next = data.get('next', '')
            link_prev = data.get('previous', '')
            data = data['results']
        if isinstance(data, (ReturnList, ReturnDict)):
            ser = data.serializer
            data = dict(data=data, serializer=ser.child if isinstance(ser, ListSerializer) else ser,
                        link_next=link_next, link_prev=link_prev)

            if getattr(ser, '_errors', {}):
                # CAUTION: bad hacks start here. This should be removed at some point

                # unmark exception from response because this was a validation error
                # This will allow TemplateHTMLRenderer to still render the template as if it was without problems
                # If this is not done, the only result user will see will be a 404 error without any details
                response = renderer_context['response']
                response.exception = False

                # result data should be object data otherwise nothing will render...
                data['data'] = data['serializer'].data

        return super().render(data, accepted_media_type, renderer_context)

    def get_template_names(self, response, view):
        if view.render_type == 'page' and DYNAMICFORMS.page_template:
            return [DYNAMICFORMS.page_template]
        return super().get_template_names(response, view)

    def get_template_context(self, data, renderer_context):
        res = super().get_template_context(data, renderer_context)
        view = renderer_context['view']
        if hasattr(view, 'template_context'):
            if callable(view.template_context):
                res.update(view.template_context())
            else:
                res.update(view.template_context)
        res['df_render_type'] = view.render_type  # This one should not fail because it's set in initialize_request
        res['DYNAMICFORMS'] = DYNAMICFORMS
        return res


# noinspection PyRedeclaration
class HTMLFormRenderer(HTMLFormRenderer):
    """
    An HTML renderer for use with templates.

    The data supplied to the Response object should be a dictionary that will
    be used as context for the template.

    The template name is determined by (in order of preference):

    1. An explicit `.template_name` attribute set on the response.
    2. An explicit `.template_name` attribute set on this class.
    3. The return result of calling `view.get_template_names()`.

    For example:
        data = {'users': User.objects.all()}
        return Response(data, template_name='users.html')
    """

    def render_field(self, field, parent_style):
        # noinspection PyProtectedMember
        if isinstance(field._field, HiddenField):
            return ''

        if isinstance(field._field, BooleanField):
            # allow also null to have max values available for determening the right value to render
            allow_null = copy.copy(field._field.allow_null)
            field._field.allow_null = True
            field.value = field._field.to_internal_value(field.value)
            field._field.allow_null = allow_null

        style = dict(self.default_style[field])
        style.update(field.style)
        if 'template_pack' not in style:
            style['template_pack'] = parent_style.get('template_pack', self.template_pack)
        style['serializer'] = parent_style.get('serializer', None)
        style['renderer'] = self

        # Get a clone of the field with text-only value representation.
        field = field.as_form_field()

        if style.get('input_type') == 'datetime-local' and isinstance(field.value, six.text_type):
            field.value = field.value.rstrip('Z')

        if 'template' in style:
            template_name = style['template']
        else:
            template_name = style['template_pack'].strip('/') + '/' + style['base_template']

        template = loader.get_template(template_name)
        context = {
            'field': field,
            'style': style,
            'DYNAMICFORMS': DYNAMICFORMS,
        }
        try:
            context.update(style['serializer'].context['view'].template_context)
        except:
            pass
        return template.render(context)

    # Jure: had to copy this one over to support custom template as parameter + DF context variable
    def render(self, data, accepted_media_type=None, renderer_context=None):
        """
        Render serializer data and return an HTML form, as a string.
        """
        renderer_context = renderer_context or {}
        form = data.serializer

        style = renderer_context.get('style', {})
        style['template_pack'] = DYNAMICFORMS.template + 'field'
        style['renderer'] = self

        template_pack = style['template_pack'].strip('/')

        # getting the template to use
        template_name = next((x for x in (  # Get the first specified template name as encountered
            style.get('form_template', None),  # template name from tag parameter
            getattr(self, 'form_template', None),  # if ViewSet set the form_template member of this renderer
            getattr(form, 'form_template', None),  # if Serializer has form_template member set
            template_pack + '/' + self.base_template  # take default template from pack
        ) if x))

        template = loader.get_template(template_name)
        context = {
            'form': form,
            'style': style,
            'DYNAMICFORMS': DYNAMICFORMS,
        }
        try:
            context.update(form.context['view'].template_context)
        except:
            pass
        return template.render(context)



import json as jsonlib

from django import template
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _
from rest_framework.templatetags import rest_framework as drftt
from rest_framework.utils.encoders import JSONEncoder

from ..action import TablePosition
from ..mixins import DisplayMode
from ..renderers import HTMLFormRenderer
from ..struct import Struct

register = template.Library()


@register.filter(name='dict_item')
def dict_item(d, k):
    """Returns the given key from a dictionary."""
    return d[k]


# just copy the template tags that are the same as we're not redeclaring them
# TODO: the following DRF filters pending removal?
# note that simply copying the filters didn't work: django kept complaining about parameters. So now it's verbatim copy
@register.filter
def items(value):
    """
    Simple filter to return the items of the dict. Useful when the dict may
    have a key 'items' which is resolved first in Django template dot-notation
    lookup. See DRF issue #4931
    Also see: https://stackoverflow.com/questions/15416662/django-template-loop-over-dictionary-items-with-items-as-key
    """
    return drftt.items(value)


@register.filter
def add_nested_class(value):
    return drftt.add_nested_class(value)


@register.filter
def format_value(value):
    # TODO: this one pending removal on tasks #62 in #63 completion
    return drftt.format_value(value)


@register.filter
def as_string(value):
    # TODO: this one pending removal on tasks #62 in #63 completion
    return drftt.as_string(value)


@register.filter
def as_list_of_strings(value):
    # TODO: this one pending removal on tasks #62 in #63 completion
    return drftt.as_list_of_strings(value)


# // just copy tags


@register.simple_tag
def render_form(serializer, form_template=None):
    """
    Renders form from serializer. If form_template is given, then renderer will use that one, otherwise it will use what
    is defined in self.base_template (e.g.: »form.html«) from template_pack (e.g.: »dynamicforms/bootstrap/field/«)

    .. code-block:: django

       {% set_var template_pack=DYNAMICFORMS.template|add:'field' %}
       {% render_form serializer template_pack=template_pack %}

    :param serializer: Serializer object
    :param form_template: form template file name to use. overrides all other template name declarations
    :return: rendered template
    """
    style = {}
    if form_template:
        style['form_template'] = form_template
    style['serializer'] = serializer

    renderer = HTMLFormRenderer()
    return renderer.render(serializer.data, None, {'style': style})


@register.simple_tag
def render_field(field, style):
    """
    Renders separate field. Style is a dict, that contains template_pack or form_template and rendered. It is defined
    when rendering form, so it is best to use that one. To do that dynamicforms should be loaded first.

    See example below.

    .. code-block:: django

       {% load dynamicforms %}
       {% for field in form %}
         {% if not field.read_only %}
           {% render_field field style=style %}
         {% endif %}
       {% endfor %}

    :param field: Serializer Field instance
    :param style: render parameters
    :return: rendered field template
    """
    renderer = style.get('renderer', HTMLFormRenderer())
    return renderer.render_field(field, style)


@register.simple_tag
def table_columns_count(serializer):
    """
    Returns number of columns including control columns

    :param serializer: Serializer
    :return: Number of all columns
    """
    actions = serializer.actions.renderable_actions(serializer)

    return (len([f for f in serializer.fields.values() if f.display_table == DisplayMode.FULL])
            + (1 if any(action.position == "rowend" for action in actions) else 0)
            + (1 if any(action.position == "rowstart" for action in actions) else 0))


@register.simple_tag(takes_context=True)
def render_table_commands(context, serializer, position, field_name=None, button_position=None):
    """
    Renders commands that are defined in serializers controls attribute.

    :param context: Context
    :param serializer: Serializer
    :param position: Position of command (See action.py->Action for more details)
    :param field_name: If position is left or right to the field, then this parameter must contain field name
    :param button_position: form button position one of (form, dialog, user-provided)
    :return: rendered command buttons. If table_header parameter is given and commands for position are defined,
        returns only rendered table header
    """
    ret = ''

    if position == 'onfieldchange':
        ret = serializer.actions.render_field_onchange(serializer)
        ret += serializer.actions.render_form_init(serializer)
    elif position == 'onfieldinit':
        ret += serializer.actions.render_field_init(serializer, field_name)
    elif position == 'form-buttons':
        ret += serializer.actions.render_form_buttons(serializer, button_position)
    else:
        table_header = None
        if position.startswith('thead_'):
            position = position[6:]
            table_header = _('Actions')

        positions = dict(
            onrowclick=(TablePosition.ROW_CLICK, TablePosition.ROW_RIGHTCLICK),
            header=(TablePosition.HEADER,),
            rowstart=(TablePosition.ROW_START,),
            rowend=(TablePosition.ROW_END,),
            fieldleft=(TablePosition.FIELD_START,),
            fieldright=(TablePosition.FIELD_END,),
        )
        ret_tmp = serializer.actions.render_renderable_actions(positions[position], field_name, serializer)

        if ret_tmp and positions[position][0] in (TablePosition.ROW_START, TablePosition.ROW_END):
            ret += ('<th>%s</th>' % table_header) if table_header else ('<td>%s</td>' % ret_tmp)
        elif ret_tmp:
            ret += ret_tmp

    ret = '{% load static i18n %}' + ret

    return mark_safe(context.template.engine.from_string(ret).render(context))


@register.simple_tag(takes_context=True)
def field_to_serializer_and_data(context, serializer):
    """
    Adjusts context such that the nested serializer field becomes THE serializer

    :param serializer: field to be converted into context
    :return: nothing
    """
    if hasattr(serializer, 'child'):
        # this is a ListSerializer
        context.update(dict(serializer=serializer.child, data=serializer.value, **serializer.child.template_context))
    else:
        serializer._field._data = serializer.value  # The serializer has default values here
        context.update(dict(serializer=serializer, data=serializer.value, **serializer.template_context))
    return ''


@register.simple_tag(takes_context=True)
def get_data_template(context, context_data=None):
    """
    Returns template that should be used for rendering current serializer data

    :param context: template context (automatically provided by django)
    :return: template file name
    """
    if context_data:
        context = context_data
    serializer = context['serializer']
    return serializer.data_template


@register.simple_tag(takes_context=True)
def set_var(context, **kwds):
    """
    Sets the given variables to provided values. Kind of like the 'with' block, only it isn't a block tag

    :param context: template context (automatically provided by django)
    :param kwds: named parameters with their respective values
    :return: this tag doesn't render
    """
    for k, v in kwds.items():
        context[k] = v
    return ''


@register.simple_tag(takes_context=True)
def set_var_conditional(context, condition=None, condition_var=None, compare=None, else_value=None, **kwds):
    """
    Sets the given variables to provided values. Kind of like the 'with' block, only it isn't a block tag

    :param context: template context (automatically provided by django)
    :param kwds: named parameters with their respective values
    :param condition_var: pair with compare to obtain True or False whether to use original assignment or else_value
    :param compare: pair with condition_var to obtain True or False whether to use original assignment or else_value
    :param condition: alternative to condition_var & compare: original assignment if truthy or else_value if falsy
    :param else_value: value to be assigned to the variable(s) when condition is falsy
    :return: this tag doesn't render
    """
    if condition_var is not None:
        condition = condition_var == compare
    for k, v in kwds.items():
        context[k] = v if condition else else_value
    return ''


@register.simple_tag(takes_context=False)
def iter_options_bound(field):
    return field.iter_options_bound(field.value)


@register.filter
def json(value):
    """
    JSON serialises given variable. Use when you want to insert a variable directly into JavaScript

    :param value: variable to serialise
    :return: JSON serialised string
    """

    class Encoder(JSONEncoder):
        def default(self, obj):
            if isinstance(obj, Struct):
                return obj.__to_dict__()
            return super().default(obj)

    return mark_safe(jsonlib.dumps(value, cls=Encoder))


@register.simple_tag(takes_context=True)
def dict_item_default(context, var, d, k, default):
    context[var] = d.get(k, default)

    return ''


@register.filter
def class_name(value):
    return value.__class__.__name__


@register.filter
def in_list(value, the_list):
    value = str(value)
    return value in the_list.split(',')


@register.filter
def startswith(text, starts):
    try:
        return text.startswith(starts)
    except:
        return False
